from typing import Union
import re
from .cache import Cache, CacheItem
from ._function_cache import CacheCall
from ._providers_caches import CACHE_ALLOWED_KWARGS, Cacheable, UnversionedProviderCache
from ._instance import _cache
import logging

log = logging.getLogger(__name__)


def cache_len():
    """Return the number of items in the cache"""
    return len(_cache)


def cache_disk_size():
    """Return the size of the cache on disk"""
    return _cache.disk_size()


def stats():
    return _cache.stats()


def entries():
    """Return all cache entries as a list of keys

    Returns
    -------
    list
        A list of all cache keys
    """
    return _cache.keys()


def add_item(key, item, expires=None):
    """Add an item to the cache with an optional expiration time in seconds

    Parameters
    ----------
    key : str
        The key under which the item will be stored
    item : any
        The item to be stored
    expires : int, optional
        The expiration time in seconds, by default None (no expiration)
    """
    _cache.set(key, item, expires)


def drop_item(key):
    """Drop an item from the cache by key

    Parameters
    ----------
    key : str
        The key of the item to be dropped
    """
    _cache.drop(key)


def get_item(key, default_value=None):
    """Get an item from the cache by key

    Parameters
    ----------
    key : str
        The key of the item to be retrieved
    default_value : any, optional
        The default value to return if the key does not exist, by default None

    Returns
    -------
    any
        The item stored under the key or the default value if the key does not exist
    """
    return _cache.get(key, default_value)


def drop_matching_entries(pattern: Union[str, re.Pattern]):
    """Drop all cache entries that match a given pattern

    Parameters
    ----------
    pattern : str or re.Pattern
        The pattern to match cache keys against
    """
    if isinstance(pattern, str):
        pattern = re.compile(pattern)
    for key in filter(pattern.match, _cache.keys()):
        log.debug(f"Dropping cache entry {key}")
        _cache.drop(key)



import os
import tarfile
from glob import glob
from tempfile import TemporaryDirectory

from speasy.core import http, any_files
from ._cdf_masters_parser import update_tree
from ._xml_catalogs_parser import load_xml_catalog
from ....config import cdaweb as cda_cfg
from ....core.index import index
from ....core.inventory.indexes import SpeasyIndex, to_dict, from_dict

_MASTERS_CDF_PATH = f"{cda_cfg.inventory_data_path()}/masters_cdf/"
_XML_CATALOG_PATH = f"{cda_cfg.inventory_data_path()}/all.xml"

_CDAWEB_INVENTORY_ = "cdaweb-inventory"
_CDAWEB_INVENTORY_TREE_ = "tree_v2"
_CDAWEB_INVENTORY_LAST_MODIFIED_MASTERS_ = "masters-last-modified"
_CDAWEB_INVENTORY_LAST_MODIFIED_XML_ = "last_modified_xml"


def _ensure_path_exists(path: str):
    dirname = os.path.dirname(path)
    os.makedirs(dirname, exist_ok=True)


def _clean_master_cdf_folder():
    _ensure_path_exists(_MASTERS_CDF_PATH)
    cdf_files = glob(f"{_MASTERS_CDF_PATH}/*.cdf")
    for cdf_file in cdf_files:
        os.remove(cdf_file)


def _download_and_extract_master_cdf(masters_url: str):
    with TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/masters.tar", 'wb') as master_archive:
            master_archive.write(any_files.any_loc_open(masters_url).read())
        tar = tarfile.open(f"{tmp_path}/masters.tar")
        tar.extractall(_MASTERS_CDF_PATH)
        tar.close()


def update_master_cdf(masters_url: str = "https://spdf.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/master.tar"):
    last_modified = http.head(masters_url).headers['last-modified']
    if index.get(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_LAST_MODIFIED_MASTERS_, "") != last_modified:
        _clean_master_cdf_folder()
        _download_and_extract_master_cdf(masters_url)
        index.set(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_LAST_MODIFIED_MASTERS_, last_modified)
        return True
    return False


def update_xml_catalog(xml_catalog_url: str = "https://spdf.gsfc.nasa.gov/pub/catalogs/all.xml"):
    last_modified = http.head(xml_catalog_url).headers['last-modified']
    if index.get(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_LAST_MODIFIED_XML_, "") != last_modified:
        _ensure_path_exists(_XML_CATALOG_PATH)
        with open(_XML_CATALOG_PATH, 'w') as f:
            f.write(http.get(xml_catalog_url).text)
            index.set(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_LAST_MODIFIED_XML_, last_modified)
            return True
    return False


def build_inventory(root: SpeasyIndex = None, xml_catalog_url: str = "https://spdf.gsfc.nasa.gov/pub/catalogs/all.xml",
                    masters_url: str = "https://spdf.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/master.tar"):
    needs_rebuild = update_xml_catalog(xml_catalog_url)
    needs_rebuild |= update_master_cdf(masters_url)
    if needs_rebuild or not index.contains(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_TREE_):
        root = load_xml_catalog(xml_file_path=_XML_CATALOG_PATH, root=root)
        update_tree(root=root, master_cdf_dir=_MASTERS_CDF_PATH)
        index.set(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_TREE_, to_dict(root, version=2))
    else:
        t = from_dict(index.get(_CDAWEB_INVENTORY_, _CDAWEB_INVENTORY_TREE_), version=2)
        root.__dict__ = t.__dict__
    return root



"""AMDA_Webservice utility functions. This module defines some conversion functions specific to AMDA_Webservice, mainly
conversion procedures for parsing CSV and VOTable data.

"""
import datetime
import logging
import os
from typing import List

from speasy.core.any_files import any_loc_open
from speasy.core.datetime_range import DateTimeRange
from speasy.products.catalog import Catalog, Event
from speasy.products.timetable import TimeTable


log = logging.getLogger(__name__)

tt_catalog_time_format = "%Y-%m-%dT%H:%M:%S.%f"


def _build_event(data, col_names: List[str]) -> Event:
    return Event(datetime.datetime.strptime(data[0], tt_catalog_time_format),
                 datetime.datetime.strptime(data[1], tt_catalog_time_format),
                 {name: value for name, value in zip(col_names[2:], data[2:])})


def load_timetable(filename: str) -> TimeTable:
    """Load a timetable file

    Parameters
    ----------
    filename: str
        filename

    Returns
    -------
    TimeTable
        File content loaded as TimeTable
    """
    if '://' not in filename:
        filename = f"file://{os.path.abspath(filename)}"
    with any_loc_open(filename) as votable:
        # save the timetable as a dataframe, speasy.common.SpeasyVariable
        # get header data first

        from astropy.io.votable import parse as parse_votable
        votable = parse_votable(votable)
        if votable.description:
            name = next(filter(lambda e: 'Name' in e,
                               votable.description.split(';\n'))).split(':')[-1]
        else:
            name = os.path.basename(filename)
        # convert astropy votable structure to SpeasyVariable
        tab = votable.get_first_table()
        # prepare data
        data = tab.array.tolist()
        dt_ranges = [DateTimeRange(datetime.datetime.strptime(t0, tt_catalog_time_format),
                                   datetime.datetime.strptime(t1, tt_catalog_time_format)) for (t0, t1) in
                     data]
        var = TimeTable(name=name, meta={}, dt_ranges=dt_ranges)
        return var


def load_catalog(filename: str) -> Catalog:
    """Load a timetable file

    Parameters
    ----------
    filename: str
        filename

    Returns
    -------
    Catalog
        File content loaded as Catalog

    """
    if '://' not in filename:
        filename = f"file://{os.path.abspath(filename)}"
    with any_loc_open(filename) as votable:
        # save the timetable as a dataframe, speasy.common.SpeasyVariable
        # get header data first

        from astropy.io.votable import parse as parse_votable
        votable = parse_votable(votable)
        # convert astropy votable structure to SpeasyVariable
        tab = votable.get_first_table()
        name = next(filter(lambda e: 'Name' in e,
                           votable.description.split(';\n'))).split(':')[-1]
        col_names = list(map(lambda f: f.name, tab.fields))
        data = tab.array.tolist()
        events = [_build_event(line, col_names) for line in data]
        var = Catalog(name=name, meta={}, events=events)
        return var


def is_public(node):
    return node.__dict__.get('is_public', True)


def is_private(node):
    return not is_public(node)



import diskcache as dc

from speasy.config import index as index_cfg


class SpeasyIndex:
    def __init__(self):
        self._index = dc.Index(index_cfg.path())

    def get(self, module, key, default=None):
        return self._index.get(f'{module}/{key}', default)

    def set(self, module, key, value):
        self._index[f'{module}/{key}'] = value

    def pop(self, module, key):
        return self._index.pop(f'{module}/{key}')

    def contains(self, module, key):
        return f'{module}/{key}' in self._index



from typing import List, Tuple, Any, Sequence, Union, Optional


def span_ctor(span_type, start, stop):
    if start <= stop:
        if span_type is list or span_type is tuple:
            return span_type((start, stop))
        else:
            return span_type(start, stop)
    else:
        return None


def is_span(maybe_span: Any) -> bool:
    return hasattr(maybe_span, '__getitem__') and len(maybe_span) == 2 and maybe_span[0] <= maybe_span[1]


def is_empty(span: Sequence) -> bool:
    return span[0] == span[1]


def intersection(span: Sequence, other: Sequence) -> Optional[Sequence]:
    if not is_span(span) or not is_span(other):
        raise TypeError("You must provide a Span like object")
    if other[0] > span[1] or span[0] > other[1]:
        return None
    else:
        return span_ctor(type(span), max(span[0], other[0]), min(span[1], other[1]))


def intersects(span: Sequence, other: Sequence) -> bool:
    return intersection(span, other) is not None


def contains(span: Sequence, other: Sequence) -> bool:
    if not is_span(span) or not is_span(other):
        raise TypeError("You must provide a Span like object")
    return span[0] <= other[0] <= span[1] and span[1] >= other[1] >= span[0]


def equals(span: Sequence, other: Sequence) -> bool:
    if not is_span(span) or not is_span(other):
        raise TypeError("You must provide a Span like object")
    return span[0] == other[0] and span[1] == other[1]


def merge(spans: Union[List[Sequence], Tuple[Sequence]]) -> List[Sequence]:
    assert all([is_span(span) for span in spans])
    merged_list = []
    spans.sort(key=lambda item: item[0])
    while len(spans):
        current_span = spans.pop(0)
        while len(spans):
            if current_span[1] >= spans[0][0]:
                current_span[1] = max(current_span[1], spans[0][1])
                spans.pop(0)
            else:
                break
        merged_list.append(current_span)
    return merged_list


def difference(span: Sequence, other_s: Sequence) -> List[Sequence]:
    if not is_span(span) or not is_span(other_s):
        raise TypeError("You must provide a Span like object")

    diff = [
        span_ctor(type(span), span[0], min(other_s[0], span[1])),
        span_ctor(type(span), max(other_s[1], span[0]), span[1])
    ]
    diff = [part for part in diff if part is not None and not is_empty(part)]
    return diff


def zoom(span: Sequence, factor: float):
    if not is_span(span):
        raise TypeError("You must provide a Span like object")
    if type(factor) is not float and type(factor) is not int:
        raise TypeError("factor must be float")

    width = (span[1] - span[0]) / 2
    center = span[0] + width
    width *= factor
    return span_ctor(type(span), center - width, center + width)


def shift(span: Sequence, distance):
    if not is_span(span):
        raise TypeError("You must provide a Span like object")
    return span_ctor(type(span), span[0] + distance, span[1] + distance)


def split(span: Sequence, fragment_duration) -> List[Sequence]:
    from math import ceil
    if not is_span(span):
        raise TypeError("You must provide a Span like object")
    total_duration = span[1] - span[0]
    if total_duration <= fragment_duration:
        return [span]
    fragments = [
        span_ctor(type(span), span[0] + fragment_duration * i, min(span[0] + fragment_duration * (i + 1), span[1])) for
        i in range(ceil(total_duration / fragment_duration))]
    return fragments



"""
.. testsetup:: *

   from speasy.core import *
   import numpy as np
"""

import os
import warnings
from collections.abc import Iterable
from datetime import datetime, timezone
from functools import wraps
from typing import Any, Dict, List, Sequence, Type, Union, Callable

import numpy as np
from dateutil.parser import parse
from tqdm.auto import tqdm

AnyDateTimeType = Union[str, datetime, np.float64, float, np.datetime64]


def deprecation(message: str) -> None:
    """Shows a deprecation warning.

    Parameters
    ----------
    message: str
        Custom message to show
    """

    warnings.warn(message, DeprecationWarning, stacklevel=2)


def pack_kwargs(**kwargs: Any) -> Dict:
    """Packs given keyword arguments into a dictionary

    Parameters
    ----------
    kwargs: Any
        Any keyword argument is accepted

    Returns
    -------
    dict
        A dict with all kwargs packed

    Examples
    --------
    >>> pack_kwargs(a=1, b="2")
    {'a': 1, 'b': '2'}
    """
    return kwargs


def all_of_type(collection: Sequence, expected_type: Type) -> bool:
    """Returns true only if the type of all elements in given collection is expected_type

    Parameters
    ----------
    collection: Sequence
        Any iterable object
    expected_type: Type
        the type you expect to match

    Returns
    -------
    bool
        True only if the type of all elements in given collection is expected_type

    Examples
    --------
    >>> all_of_type([1,2,3], int)
    True

    >>> all_of_type([1,2,3.], int)
    False
    """
    return all(map(lambda x: type(x) is expected_type, collection))


def is_collection(value: Any) -> bool:
    """

    Parameters
    ----------
    value : Any

    Returns
    -------
    bool
        True if given value is collection like object but not a string

    """
    return isinstance(value, Iterable) and type(value) is not str


def mkdir(directory: str) -> None:
    """Creates directory and parents if they do not exist

    Parameters
    ----------
    directory: str
        Path to create
    """
    os.makedirs(directory, exist_ok=True)


def listify(obj: Any) -> List:
    """Wraps inside a list anything that is not a list. Useful in for loops when you can't be sure
    the object you want to iterate is a list.

    Parameters
    ----------
    obj: Any
        Any object or list

    Returns
    -------
    list
        list(obj) if obj is not a list

    Examples
    --------
    >>> for i in listify(1):
    ...     print(i)
    ...
    1

    >>> for i in listify([1,2,3]):
    ...     print(i)
    ...
    1
    2
    3
    """
    obj_t = type(obj)
    if obj_t is list:
        return obj
    if obj_t is tuple:
        return list(obj)
    else:
        return [obj]


def make_utc_datetime(input_dt: AnyDateTimeType) -> datetime:
    """Makes UTC datetime from given input.

    Parameters
    ----------
    input_dt: str or datetime or np.float64 or float
        Datetime to convert, can be either en Epoch, a datetime or a string

    Returns
    -------
    datetime
        A datetime.datetime object forced to UTC time zone

    Examples
    --------
    >>> make_utc_datetime('2018-01-02')
    datetime.datetime(2018, 1, 2, 0, 0, tzinfo=datetime.timezone.utc)

    >>> make_utc_datetime(0.)
    datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)

    >>> from datetime import datetime
    >>> make_utc_datetime(datetime(2020,1,1))
    datetime.datetime(2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    if type(input_dt) in (np.float64, float):
        return datetime.fromtimestamp(input_dt, tz=timezone.utc)
    if type(input_dt) is str:
        input_dt = parse(input_dt)
    if type(input_dt) is np.datetime64:
        if input_dt.dtype == np.dtype('datetime64[ns]'):
            return datetime.fromtimestamp(input_dt.astype(np.int64) * 1e-9, tz=timezone.utc)

    return datetime(input_dt.year, input_dt.month, input_dt.day, input_dt.hour, input_dt.minute, input_dt.second,
                    input_dt.microsecond, tzinfo=timezone.utc)


def make_utc_datetime64(input_dt: AnyDateTimeType) -> np.datetime64:
    """Makes UTC np.datetime64 from given input.

    Parameters
    ----------
    input_dt: str or datetime or np.float64 or float
        Datetime to convert, can be either en Epoch, a datetime or a string

    Returns
    -------
    np.datetime64
        A numpy datetime64 object forced to UTC time zone

    Examples
    --------
    >>> make_utc_datetime64('2018-01-02')
    np.datetime64('2018-01-02T00:00:00.000000000')

    >>> make_utc_datetime64(0.)
    np.datetime64('1970-01-01T00:00:00.000000')

    >>> from datetime import datetime
    >>> make_utc_datetime64(datetime(2020,1,1))
    np.datetime64('2020-01-01T00:00:00.000000000')
    """
    if type(input_dt) in (np.float64, float):
        return np.datetime64(datetime.fromtimestamp(input_dt, tz=timezone.utc))
    if type(input_dt) is str:
        input_dt = parse(input_dt)
    if type(input_dt) is np.datetime64:
        if input_dt.dtype == np.dtype('datetime64[ns]'):
            return input_dt

    return np.datetime64(
        datetime(input_dt.year, input_dt.month, input_dt.day, input_dt.hour, input_dt.minute, input_dt.second,
                 input_dt.microsecond, tzinfo=timezone.utc), 'ns')


def epoch_to_datetime64(epoch_array: np.array) -> np.array:
    """Converts an array of floats encoded as Unix Epoch (seconds since 1970) to an array of numpy datetime64[ns]

    Parameters
    ----------
    epoch_array : np.array
        Input array of folats (Epoch)

    Returns
    -------
    np.array
        Output array of datetime64[ns]

    Examples
    --------
    >>> epoch_to_datetime64(np.arange(2))
    array(['1970-01-01T00:00:00.000000000', '1970-01-01T00:00:01.000000000'],
          dtype='datetime64[ns]')


    """
    return (epoch_array * 1e9).astype("datetime64[ns]")


def datetime64_to_epoch(datetime64_array: np.array) -> np.array:
    """Converts an array of numpy datetime64[ns] to an array of floats encoded as Unix Epoch (seconds since 1970)

    Parameters
    ----------
    datetime64_array : np.array
        Input array of datetime64[ns]

    Returns
    -------
    np.array
        Output array of floats (Epoch)

    Examples
    --------
    >>> datetime64_to_epoch(np.array(['1970-01-01T00:00:00.000000000', '1970-01-01T00:00:01.000000000'],
    ...                              dtype='datetime64[ns]'))
    array([0., 1.])
    """
    return (datetime64_array.astype("int64") * 1e-9).astype("float64")


class AllowedKwargs(object):
    """A decorator that prevent from passing unexpected kwargs to a function
    """

    def __init__(self, allowed_list):
        self.allowed_list = set(allowed_list)

    def __call__(self, func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            unexpected_args = list(
                filter(lambda arg_name: arg_name not in self.allowed_list, kwargs.keys()))
            if not unexpected_args:
                return func(*args, **kwargs)
            raise TypeError(
                f"Unexpected keyword argument {unexpected_args}, allowed keyword arguments are {self.allowed_list}")

        return wrapped


def fix_name(name: str):
    """Makes given input compatible with python charset https://docs.python.org/3/reference/lexical_analysis.html#identifiers

    Parameters
    ----------
    name: str
        input string to sanitize

    Returns
    -------
    str
        a string compatible with python naming rules


    Examples
    --------
    >>> fix_name('Parker Solar Probe (PSP)')
    'Parker_Solar_Probe_PSP'

    >>> fix_name('IS⊙ISEPI_Lo')
    'ISoISEPI_Lo'

    >>> fix_name('all_Legal_strings_123')
    'all_Legal_strings_123'

    """
    rules = (
        ('-', '_'),
        (':', '_'),
        ('.', '_'),
        ('(', ''),
        (')', ''),
        ('/', ''),
        (' ', '_'),
        ('{', ''),
        ('}', ''),
        ('(', ''),
        ('⊙', 'o'),
        (';', '_'),
        (',', '_')
    )
    if len(name):
        if name[0].isnumeric():
            name = "n_" + name
        for bad, replacement in rules:
            if bad in name:
                name = name.replace(bad, replacement)
        return name
    raise ValueError("Got empty name")


def progress_bar(leave=True, progress=False, desc=None, **kwargs):
    if not progress:
        return lambda x: x
    else:
        return lambda x: tqdm(x, leave=leave, desc=desc)


class EnsureUTCDateTime(object):

    def __call__(self, get_data: Callable):
        @wraps(get_data)
        def wrapped(wrapped_self, product, start_time, stop_time, **kwargs):
            return get_data(wrapped_self, product=product, start_time=make_utc_datetime(start_time),
                            stop_time=make_utc_datetime(stop_time), **kwargs)

        return wrapped



from typing import Dict, Tuple, Callable
from urllib.parse import urlparse, urlencode
from functools import wraps

from speasy.config import core as core_config

_REWRITE_RULES_ = core_config.http_rewrite_rules.get()


def quote(*args, **kwargs):
    from urllib.parse import quote as _quote
    return _quote(*args, **kwargs)


def ensure_url_scheme(url: str) -> str:
    """Adds file:// to url for local files

    Parameters
    ----------
    url : str
        url with or without scheme
    Returns
    -------
    str
        url with 'file:' scheme added when none was provided else input url
    """
    parsed = urlparse(url)
    if parsed.scheme != 'file' and parsed.netloc == '':
        return f"file://{url}"
    return url


def is_local_file(url: str):
    """Returns true if url correspond to a local path.

    Parameters
    ----------
    url : str
        file url formatted as local path or standard URL format (https://en.wikipedia.org/wiki/URL)

    Returns
    -------
    bool
        True if matches any local path/URL


    Examples
    --------

    >>> from speasy.core.url_utils import is_local_file

    >>> is_local_file("/some/path")
    True

    >>> is_local_file("C:/some/path")
    True

    >>> is_local_file("file:///some/path")
    True

    >>> is_local_file("http://some/path")
    False

    """
    split_url = urlparse(url)
    return split_url.scheme in ('', 'file') or split_url.netloc == ''


def build_url(base: str, parameters: Dict) -> str:
    return base + '?' + urlencode(parameters)


def host_and_port(url: str) -> Tuple[str, int]:
    """Returns the host and port of an url

    Parameters
    ----------
    url : str
        url with or without scheme

    Returns
    -------
    Tuple[str, str]
        host and port of the url, defaults to 80 if no port or scheme is provided
    """
    parsed = urlparse(url)
    if parsed.port is not None:
        return parsed.hostname, parsed.port
    elif parsed.scheme == 'http':
        return parsed.hostname, 80
    elif parsed.scheme == 'https':
        return parsed.hostname, 443

    return parsed.hostname, 80


def apply_rewrite_rules(url: str) -> str:
    for base_url in _REWRITE_RULES_:
        if url.startswith(base_url):
            return _REWRITE_RULES_[base_url] + url[len(base_url):]
    return url


class ApplyRewriteRules:

    def __init__(self, is_method: bool = False):
        self.is_method = is_method

    def __call__(self, f: Callable):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if self.is_method:
                wrapped_self = args[0]
                args = args[1:]
            if 'url' in kwargs:
                url = kwargs.pop('url')
            elif len(args):
                url = args[0]
                args = args[1:]
            else:
                url = None
            if url is not None:
                url = apply_rewrite_rules(url)
            if self.is_method:
                return f(wrapped_self, url, *args, **kwargs)
            return f(url, *args, **kwargs)

        return wrapper


def extract_path(url: str) -> str:
    return urlparse(url).path



class SpeasyProduct(object):
    __slots__ = ['__request_params']

    def __init__(self, request_params=None):
        self.__request_params = request_params

    @property
    def request_params(self):
        return self.__request_params



import io
import logging
import tarfile
from datetime import datetime, timedelta
from tempfile import TemporaryDirectory
from typing import Optional, Tuple, Dict

import numpy as np
from astroquery.utils.tap.core import TapPlus

from speasy.core import any_files, AllowedKwargs, fix_name, EnsureUTCDateTime
from speasy.core.codecs import get_codec, CodecInterface
from speasy.core.cache import Cacheable, CACHE_ALLOWED_KWARGS  # _cache is used for tests (hack...)
from speasy.core.dataprovider import DataProvider, ParameterRangeCheck, GET_DATA_ALLOWED_KWARGS
from speasy.core.datetime_range import DateTimeRange
from speasy.core.inventory.indexes import ParameterIndex, DatasetIndex, SpeasyIndex, make_inventory_node
from speasy.core.proxy import Proxyfiable, GetProduct, PROXY_ALLOWED_KWARGS
from speasy.core.requests_scheduling import SplitLargeRequests
from speasy.core.url_utils import build_url
from speasy.products.variable import SpeasyVariable

log = logging.getLogger(__name__)


def _only_primitive_types(d: dict) -> dict:
    for k, v in d.items():
        if not isinstance(v, (int, bool, str, type(None), list, tuple, set)):
            if isinstance(v, np.integer):
                d[k] = int(v)
            elif isinstance(v, np.floating):
                d[k] = float(v)
            elif isinstance(v, (np.bool_,bool)):
                d[k] = bool(v)
    return d


def to_dataset_and_variable(index_or_str: ParameterIndex or str) -> Tuple[str, str]:
    if type(index_or_str) is str:
        parts = index_or_str.split('/')
    elif isinstance(index_or_str, ParameterIndex):
        parts = index_or_str.product.split('/')
    else:
        raise TypeError(f"given parameter {index_or_str} of type {type(index_or_str)} is not a compatible index")
    assert len(parts) == 2
    return parts[0], parts[1]


def register_dataset(instruments, datasets, dataset):
    meta = {cname: dataset[cname] for cname in dataset.colnames}
    meta['stop_date'] = meta.pop('end_date')
    name = fix_name(meta['dataset_id'])
    node = make_inventory_node(instruments[dataset['instruments']], DatasetIndex, name=name,
                               provider="csa",
                               uid=meta['dataset_id'], **_only_primitive_types(meta))
    datasets[meta['dataset_id']] = node


def register_observatory(missions, observatories, observatory):
    meta = {cname: observatory[cname] for cname in observatory.colnames}
    name = meta.pop('name')
    node = make_inventory_node(missions[observatory['mission_name']], SpeasyIndex,
                               name=fix_name(name),
                               provider="csa",
                               uid=name,
                               **_only_primitive_types(meta))
    observatories[name] = node


def register_mission(inventory_tree, missions, mission):
    meta = {cname: mission[cname] for cname in mission.colnames}
    name = meta.pop('name')
    node = make_inventory_node(inventory_tree, SpeasyIndex, name=fix_name(name),
                               provider="csa",
                               uid=name, **_only_primitive_types(meta))
    missions[name] = node


def register_instrument(observatories, instruments, instrument):
    meta = {cname: instrument[cname] for cname in instrument.colnames}
    name = meta.pop('name')
    node = make_inventory_node(observatories.get(instrument['observatories'], observatories['MULTIPLE']),
                               SpeasyIndex, name=fix_name(name),
                               provider="csa",
                               uid=name, **_only_primitive_types(meta))
    instruments[name] = node


def register_param(datasets, parameter):
    parent_dataset = datasets.get(parameter["dataset_id"], None)
    if parent_dataset is not None:
        meta = {cname: parameter[cname] for cname in parameter.colnames}
        meta['dataset'] = parameter["dataset_id"]
        meta['start_date'] = parent_dataset.start_date
        meta['stop_date'] = parent_dataset.stop_date
        name = fix_name(meta['parameter_id'])
        make_inventory_node(parent_dataset, ParameterIndex, name=name,
                            provider="csa", uid=f"{parameter['dataset_id']}/{parameter['parameter_id']}",
                            **_only_primitive_types(meta))


def build_inventory(root: SpeasyIndex, tapurl="https://csa.esac.esa.int/csa-sl-tap/tap/"):
    CSA = TapPlus(url=tapurl)
    missions_req = CSA.launch_job_async("SELECT * FROM csa.v_mission")
    observatories_req = CSA.launch_job_async("SELECT * FROM csa.v_observatory")
    instruments_req = CSA.launch_job_async("SELECT * FROM csa.v_instrument")
    datasets_req = CSA.launch_job_async(
        "SELECT * FROM csa.v_dataset WHERE  dataset_id like '%GRMB' OR (is_cef='true' AND is_istp='true')")
    parameters_req = CSA.launch_job_async("SELECT * FROM csa.v_parameter WHERE data_type='Data' AND value_type<>'CHAR'")
    missions = {}
    observatories = {}
    instruments = {}
    datasets = {}
    list(map(lambda m: register_mission(root, missions, m), missions_req.get_results()))
    list(map(lambda o: register_observatory(missions, observatories, o), observatories_req.get_results()))
    list(map(lambda i: register_instrument(observatories, instruments, i), instruments_req.get_results()))
    list(map(lambda d: register_dataset(instruments, datasets, d), datasets_req.get_results()))
    list(map(lambda p: register_param(datasets, p), parameters_req.get_results()))

    return root


def _load_variable(archive: io.BytesIO, variable: str, cdf_codec: CodecInterface) -> Optional[SpeasyVariable]:
    with tarfile.open(fileobj=archive) as tar:
        tarname = tar.getnames()
        if len(tarname):
            with TemporaryDirectory() as tmp_dir:
                tar.extractall(tmp_dir)
                return cdf_codec.load_variable(file=f"{tmp_dir}/{tarname[0]}", variable=variable)
    return None


def get_parameter_args(start_time: datetime, stop_time: datetime, product: str, **kwargs):
    return {'path': f"csa/{product}", 'start_time': f'{start_time.isoformat()}',
            'stop_time': f'{stop_time.isoformat()}'}


class CsaWebservice(DataProvider):
    BASE_URL = "https://csa.esac.esa.int"

    def __init__(self):
        DataProvider.__init__(self, provider_name='csa')
        self.__url = f"{self.BASE_URL}/csa-sl-tap/data"
        self._cdf_codec = get_codec("application/x-cdf")

    def _dataset_range(self, dataset: str or DatasetIndex) -> DateTimeRange:
        if type(dataset) is str:
            dataset = self.flat_inventory.datasets[dataset]
        return DateTimeRange(dataset.start_date, dataset.stop_date)

    def _dl_variable(self,
                     dataset: str, variable: str,
                     start_time: datetime, stop_time: datetime, extra_http_headers: Dict[str, str] or None = None) -> \
            Optional[SpeasyVariable]:

        # https://csa.esac.esa.int/csa-sl-tap/data?RETRIEVAL_TYPE=product&&DATASET_ID=C3_CP_PEA_LERL_DEFlux&START_DATE=2001-06-10T22:12:14Z&END_DATE=2001-06-11T06:12:14Z&DELIVERY_FORMAT=CDF_ISTP&DELIVERY_INTERVAL=all
        ds_range = self._dataset_range(dataset)
        if not ds_range.intersect(DateTimeRange(start_time, stop_time)):
            log.warning(f"You are requesting {dataset}/{variable} outside of its definition range {ds_range}")
            return None
        headers = {}
        if extra_http_headers is not None:
            headers.update(extra_http_headers)
        return _load_variable(
            any_files.any_loc_open(
                build_url(base=self.__url, parameters={
                    "RETRIEVAL_TYPE": "product",
                    "DATASET_ID": dataset,
                    "START_DATE": start_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                    "END_DATE": stop_time.strftime('%Y-%m-%dT%H:%M:%SZ'),
                    "DELIVERY_FORMAT": "CDF_ISTP",
                    "DELIVERY_INTERVAL": "all"
                }),
                headers=headers),
            variable, cdf_codec=self._cdf_codec)

    @staticmethod
    def build_inventory(root: SpeasyIndex):
        return build_inventory(root)

    def parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.csa.parameter_range("C3_CP_WBD_WAVEFORM_BM2/B__C3_CP_WBD_WAVEFORM_BM2")
        <DateTimeRange: ... -> ...>

        """
        return self._parameter_range(parameter_id)

    def dataset_range(self, dataset_id: str or DatasetIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        dataset_id: str or DatasetIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.csa.dataset_range("D2_CP_FGM_SPIN")
        <DateTimeRange: 2004-07-27T00:00:00+00:00 -> ...>

        """
        return self._dataset_range(dataset_id)

    def product_last_update(self, product: str or ParameterIndex):
        """Get date of last modification of dataset or parameter.

        Parameters
        ----------
        product: str or ParameterIndex
            product

        Returns
        -------
        str
            product last update date
        """
        dataset, _ = to_dataset_and_variable(product)
        return self.flat_inventory.datasets[dataset].date_last_update

    @AllowedKwargs(PROXY_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + GET_DATA_ALLOWED_KWARGS)
    @EnsureUTCDateTime()
    @ParameterRangeCheck()
    @Cacheable(prefix="csa", fragment_hours=lambda x: 12, version=product_last_update)
    @SplitLargeRequests(threshold=lambda x: timedelta(days=7))
    @Proxyfiable(GetProduct, get_parameter_args)
    def get_data(self, product, start_time: datetime, stop_time: datetime,
                 extra_http_headers: Dict[str, str] or None = None):
        dataset, variable = to_dataset_and_variable(product)
        return self._dl_variable(start_time=start_time, stop_time=stop_time, dataset=dataset,
                                 variable=variable, extra_http_headers=extra_http_headers)

    def get_variable(self, dataset: str, variable: str, start_time: datetime or str, stop_time: datetime or str,
                     **kwargs) -> \
            Optional[SpeasyVariable]:
        return self.get_data(f"{dataset}/{variable}", start_time, stop_time, **kwargs)



import base64
import inspect
from datetime import timedelta
from functools import wraps
from typing import Callable

from ._instance import _cache


def make_key_from_args(*args, **kwargs):
    key = list(map(str, args))
    key += list(map(lambda k: str(k) + "=" + str(kwargs[k]), sorted(kwargs.keys())))
    result = ','.join(key)
    return base64.b64encode(result.encode()).decode()


class CacheCall(object):
    def __init__(self, cache_retention=60 * 15, is_pure=False, cache_instance=_cache):
        if type(cache_retention) is timedelta:
            cache_retention = cache_retention.total_seconds()
        self.cache_retention = cache_retention
        self.cache = cache_instance
        self.is_methode = False
        self.is_pure = is_pure

    def add_to_cache(self, cache_entry, value):
        if value is not None:
            self.cache.set(cache_entry, value, expire=self.cache_retention)
        return value

    def get_from_cache(self, cache_entry):
        return self.cache.get(cache_entry, None)

    def __call__(self, function: Callable):
        spec = inspect.getfullargspec(function)
        if len(spec.args) and 'self' == spec.args[0]:
            self.is_methode = True
        if self.is_pure:
            cache_entry_prefix = f"__internal__/CacheCall/{function.__module__}/{function.__qualname__}"
        else:
            cache_entry_prefix = f"__internal__/CacheCall/{hash(function)}"

        @wraps(function)
        def wrapped(*args, disable_cache=False, force_refresh=False, **kwargs):
            args_to_hash = args[1:] if self.is_methode and self.is_pure else args
            cache_entry = cache_entry_prefix + "/" + make_key_from_args(*args_to_hash, **kwargs)
            if disable_cache:
                return function(*args, **kwargs)
            if force_refresh:
                return self.add_to_cache(cache_entry, function(*args, **kwargs))
            else:
                return self.get_from_cache(cache_entry) or self.add_to_cache(cache_entry, function(*args, **kwargs))

        return wrapped



# -*- coding: utf-8 -*-

"""CDA_Webservice package for Space Physics WebServices Client."""

__author__ = """Alexis Jeandet"""
__email__ = 'alexis.jeandet@member.fsf.org'
__version__ = '0.1.0'

import logging
import re
from datetime import datetime, timedelta
from typing import Dict, Optional, Tuple

from speasy.core import AllowedKwargs, EnsureUTCDateTime
from speasy.core import http, url_utils
from speasy.config import cdaweb as cda_cfg
from speasy.core.codecs import get_codec
from speasy.core.cache import CACHE_ALLOWED_KWARGS, UnversionedProviderCache
from speasy.core.dataprovider import (GET_DATA_ALLOWED_KWARGS, DataProvider,
                                      ParameterRangeCheck)
from speasy.core.datetime_range import DateTimeRange
from speasy.core.inventory.indexes import (DatasetIndex, ParameterIndex,
                                           SpeasyIndex)
from speasy.core.proxy import PROXY_ALLOWED_KWARGS, GetProduct, Proxyfiable
from speasy.core.requests_scheduling import SplitLargeRequests
from speasy.core.direct_archive_downloader import get_product as direct_archive_get_product
from speasy.products.variable import SpeasyVariable
from ._direct_archive import to_direct_archive_params

log = logging.getLogger(__name__)

_burst_regex = re.compile("(.*MMS.*FPI.*BRST.*|.*MMS.*SCM.*BRST.*)")


def _is_burst_product(product: ParameterIndex or str) -> bool:
    if isinstance(product, ParameterIndex):
        product = product.spz_uid()
    return bool(_burst_regex.match(str(product)))


def _is_virtual_parameter(product: ParameterIndex) -> bool:
    return product.__dict__.get('VIRTUAL', 'FALSE').upper() == 'TRUE'


def _large_request_max_duration(product):
    if _is_burst_product(product):
        return timedelta(hours=2)
    else:
        return timedelta(days=7)


def _cache_fragment_size(product):
    if _is_burst_product(product):
        return 2
    else:
        return 12


class CdaWebException(Exception):
    def __init__(self, text):
        super(CdaWebException, self).__init__(text)


def get_parameter_args_ws(start_time: datetime, stop_time: datetime, product: str, **_):
    return {
        'path': f"cdaweb/{product}",
        'start_time': f'{start_time.isoformat()}',
        'stop_time': f'{stop_time.isoformat()}',
        'method': 'API',
    }


class CdaWebservice(DataProvider):
    BASE_URL = "https://cdaweb.gsfc.nasa.gov"

    def __init__(self):
        self.__url = f"{self.BASE_URL}/WS/cdasr/1"
        DataProvider.__init__(self, provider_name='cda', provider_alt_names=['cdaweb'])
        self._cdf_codec = get_codec('application/x-cdf')

    def build_inventory(self, root: SpeasyIndex):
        from ._inventory_builder import build_inventory
        root = build_inventory(root=root)
        return root

    def parameter_range(self, parameter_id: str or ParameterIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        parameter_id: str or ParameterIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.cda.parameter_range("AC_H0_MFI/BGSEc")
        <DateTimeRange: 1997-09-02T00:00:12+00:00 -> ...>

        """
        return self._parameter_range(parameter_id)

    def dataset_range(self, dataset_id: str or DatasetIndex) -> Optional[DateTimeRange]:
        """Get product time range.

        Parameters
        ----------
        dataset_id: str or DatasetIndex
            parameter id

        Returns
        -------
        Optional[DateTimeRange]
            Data time range

        Examples
        --------

        >>> import speasy as spz
        >>> spz.cda.dataset_range("AC_H0_MFI")
        <DateTimeRange: 1997-09-02T00:00:12+00:00 -> ...>

        """
        return self._dataset_range(dataset_id)

    def _to_dataset_and_variable(self, index_or_str: ParameterIndex or str) -> Tuple[str, str]:

        if isinstance(index_or_str, ParameterIndex):
            index_or_str = index_or_str.spz_uid()

        if type(index_or_str) is str:
            if '/' in index_or_str:
                parts = index_or_str.split('/')
                if len(parts) == 2:
                    return parts[0], parts[1]
                for pos in range(1, len(parts)):
                    ds = '/'.join(parts[:pos])
                    var = '/'.join(parts[pos:])
                    if (ds in self.flat_inventory.datasets) and (index_or_str in self.flat_inventory.parameters):
                        return ds, var
                raise ValueError(
                    f"Given string is ambiguous, it contains several '/', tried all combinations but failed to find a matching dataset/variable pair in inventory: {index_or_str}")
            raise ValueError(f"Given string does not look like a CDA dataset/variable pair: {index_or_str}")
        raise TypeError(f"Wrong type for {index_or_str}, expecting a string or a SpeasyIndex, got {type(index_or_str)}")

    def _dl_variable(self,
                     dataset: str, variable: str,
                     start_time: datetime, stop_time: datetime, if_newer_than: datetime or None = None,
                     extra_http_headers: Dict or None = None) -> Optional[SpeasyVariable]:
        start_time, stop_time = start_time.strftime('%Y%m%dT%H%M%SZ'), stop_time.strftime('%Y%m%dT%H%M%SZ')
        fmt = "cdf"
        url = f"{self.__url}/dataviews/sp_phys/datasets/{url_utils.quote(dataset, safe='')}/data/{start_time},{stop_time}/{url_utils.quote(variable, safe='')}?format={fmt}"
        headers = {"Accept": "application/json"}
        if if_newer_than is not None:
            headers["If-Modified-Since"] = if_newer_than.ctime()
        if extra_http_headers is not None:
            headers.update(extra_http_headers)
        resp = http.get(url, headers=headers)
        log.debug(resp.url)
        if resp.status_code == 200 and 'FileDescription' in resp.json():
            return self._cdf_codec.load_variable(file=resp.json()['FileDescription'][0]['Name'], variable=variable)
        elif not resp.ok:
            if resp.status_code == 404 and "No data available" in resp.json().get('Message', [""])[0]:
                log.warning(f"Got 404 'No data available' from CDAWeb with {url}")
                return None
            raise CdaWebException(f'Failed to get data with request: {url}, got {resp.status_code} HTTP response')
        else:
            return None

    @UnversionedProviderCache(prefix="cda", fragment_hours=_cache_fragment_size, cache_retention=timedelta(days=7))
    @SplitLargeRequests(threshold=_large_request_max_duration)
    @Proxyfiable(GetProduct, get_parameter_args_ws)
    def _get_data_with_ws(self, product, start_time: datetime, stop_time: datetime,
                          if_newer_than: datetime or None = None,
                          extra_http_headers: Dict or None = None) -> Optional[SpeasyVariable]:
        dataset, variable = self._to_dataset_and_variable(product)
        return self._dl_variable(start_time=start_time, stop_time=stop_time, dataset=dataset,
                                 variable=variable, if_newer_than=if_newer_than, extra_http_headers=extra_http_headers)

    def _get_data_with_direct_archive(self, product, start_time: datetime, stop_time: datetime, mode_is_best: bool,
                                      if_newer_than: datetime or None = None,
                                      extra_http_headers: Dict or None = None) -> Optional[SpeasyVariable]:

        dataset, variable = self._to_dataset_and_variable(product)
        dataset = self.flat_inventory.datasets[dataset]
        if type(product) is str:
            product_index = self.flat_inventory.parameters[product]
        else:
            product_index = product
        archive_params = to_direct_archive_params(file_naming=dataset.filenaming,
                                                  subdivided_by=dataset.subdividedby,
                                                  url=dataset.url)
        log.debug(f"Trying to get {product} with direct_archive method, archive_params={archive_params}")
        if archive_params is not None and not _is_virtual_parameter(product_index):
            return direct_archive_get_product(variable=variable, start_time=start_time, stop_time=stop_time,
                                              **archive_params,
                                              master_cdf_url=dataset.mastercdf)
        else:
            if not mode_is_best:
                log.warning(f"Can't get {product} without web service, switching to web service")
            return self._get_data_with_ws(product=product, start_time=start_time, stop_time=stop_time,
                                          if_newer_than=if_newer_than, extra_http_headers=extra_http_headers)

    @AllowedKwargs(
        PROXY_ALLOWED_KWARGS + CACHE_ALLOWED_KWARGS + GET_DATA_ALLOWED_KWARGS + ['if_newer_than', 'method'])
    @EnsureUTCDateTime()
    @ParameterRangeCheck()
    def get_data(self, product, start_time: datetime, stop_time: datetime, if_newer_than: datetime or None = None,
                 extra_http_headers: Dict or None = None, method: Optional[str] = None, **kwargs) -> Optional[
        SpeasyVariable]:
        method = method or cda_cfg.preferred_access_method.get()
        if method.upper() in ('FILE', 'BEST'):
            return self._get_data_with_direct_archive(product=product, start_time=start_time, stop_time=stop_time,
                                                      if_newer_than=if_newer_than,
                                                      extra_http_headers=extra_http_headers,
                                                      mode_is_best=method == 'best')
        else:
            return self._get_data_with_ws(product=product, start_time=start_time, stop_time=stop_time,
                                          if_newer_than=if_newer_than, extra_http_headers=extra_http_headers, **kwargs)

    def get_variable(self, dataset: str, variable: str, start_time: datetime or str, stop_time: datetime or str,
                     **kwargs) -> \
        Optional[SpeasyVariable]:
        return self.get_data(f"{dataset}/{variable}", start_time, stop_time, **kwargs)



from typing import Optional, Union, Dict, Any, Tuple
import io
import logging

import numpy as np
import pandas as pds
import json

from speasy.core.codecs.codec_interface import Buffer
from speasy.core.any_files import any_loc_open
from .csv_file import HapiCsvFile

log = logging.getLogger(__name__)


def _extract_headers(file: io.IOBase) -> Dict[str, Any]:
    headers = []
    line = file.readline()
    while line:
        if line.startswith('#'):
            headers.append(line[1:].strip())
        else:
            break
        line = file.readline()
    if len(headers):
        return json.loads(''.join(headers))
    return {}


def _extract_data(file: io.IOBase) -> pds.DataFrame:
    file.seek(0)
    return pds.read_csv(file, comment='#', sep=',', header=None, skiprows=0, parse_dates=[0], index_col=0)


def _parse_hapi_csv(file: io.IOBase) -> Tuple[pds.DataFrame, Dict[str, Any]]:
    headers = _extract_headers(file)
    assert headers["parameters"][0]["type"] == "isotime"
    data = _extract_data(file)
    return data, headers


def _load_csv(file: Union[Buffer, str, io.IOBase]) -> Tuple[Optional[pds.DataFrame], Optional[Dict[str, Any]]]:
    if isinstance(file, str):
        with any_loc_open(file, cache_remote_files=False, mode='r') as f:
            return _parse_hapi_csv(f)
    if isinstance(file, io.IOBase) or hasattr(file, 'read'):
        return _parse_hapi_csv(file)
    return None, None


def load_hapi_csv(file: Union[Buffer, str, io.IOBase]) -> Optional[HapiCsvFile]:
    data, headers = _load_csv(file)
    hapi_csv_file = HapiCsvFile()
    if data is not None and headers is not None:
        time_header = headers["parameters"][0]
        assert time_header["type"] == "isotime"
        hapi_csv_file.create_parameter(data.index.to_numpy(dtype="datetime64[ns]"),
                                       meta=time_header)
        column_offset = 0
        for i, param_meta in enumerate(headers["parameters"][1:]):
            shape = param_meta.get("size", [1])
            flatten_shape = np.prod(shape)
            hapi_csv_file.create_parameter(data.iloc[:, column_offset:column_offset + flatten_shape].to_numpy().reshape(
                [len(data.index)] + shape), meta=param_meta)
            column_offset += flatten_shape
        return hapi_csv_file
    return None



import io
import logging
import os
import re
import time
from datetime import timedelta, datetime, timezone
from typing import List, Optional, Union
from threading import get_native_id
from speasy.core.cache import CacheCall
from speasy.core.cache import get_item, add_item, CacheItem
from . import http
from .url_utils import is_local_file, extract_path

log = logging.getLogger(__name__)
_HREF_REGEX = re.compile('''href=['"]([A-Za-z0-9-_./]+)['"]>''')


class PendingRequest:
    def __init__(self):
        self._start_time = datetime.now(tz=timezone.utc)
        self._pid = get_native_id()

    @property
    def elapsed_time(self):
        return datetime.now(tz=timezone.utc) - self._start_time

    @property
    def pid(self):
        return self._pid


class AnyFile(io.IOBase):
    def __init__(self, url, file_impl: io.IOBase, status=200):
        self._url = url
        self._file_impl = file_impl
        self._status = status

    @property
    def url(self):
        return self._url

    def read(self, *args, **kwargs):
        return self._file_impl.read(*args, **kwargs)

    def readline(self, *args, **kwargs):
        return self._file_impl.readline(*args, **kwargs)

    def seek(self, *args, **kwargs):
        return self._file_impl.seek(*args, **kwargs)

    def close(self):
        return self._file_impl.close()

    @property
    def ok(self):
        return (self._status in (200, 304)) and self._file_impl.readable()

    @property
    def status_code(self):
        return self._status

    def __del__(self):
        if not self._file_impl.closed:
            self.close()

    def __getattr__(self, item):
        return getattr(self._file_impl, item)


def _remote_open(url, timeout: int = http.DEFAULT_TIMEOUT, headers: dict = None, mode='rb'):
    resp = http.urlopen(url=url, headers=headers, timeout=timeout)
    if 'b' in mode:
        return AnyFile(url, io.BytesIO(resp.bytes))
    else:
        return AnyFile(url, io.StringIO(resp.text))


def _make_file_from_cache_entry(entry: CacheItem, url: str, mode: str) -> AnyFile:
    if 'b' in mode:
        return AnyFile(url, io.BytesIO(entry.data))
    else:
        return AnyFile(url, io.StringIO(entry.data))


def _wait_for_pending_request(url: str, timeout: int):
    cache_item: Optional[CacheItem] = get_item(url)
    while type(cache_item) is PendingRequest:
        if cache_item.elapsed_time > timedelta(seconds=timeout):  # pragma: no cover
            log.warning(f"Pending request for {url} timed out")
            return None
        time.sleep(.1)
        cache_item = get_item(url)
    return cache_item


def _try_lock_request(url: str, timeout: int):
    cache_item: Optional[CacheItem] = get_item(url)
    if cache_item is None:
        add_item(key=url, item=PendingRequest())
        cache_item: Optional[CacheItem] = get_item(url)
    if type(cache_item) is PendingRequest:
        if cache_item.pid != get_native_id():
            return _wait_for_pending_request(url, timeout)
    return cache_item


def _cached_get_remote_file(url, timeout: int = http.DEFAULT_TIMEOUT, headers: dict = None, mode='rb') -> AnyFile:
    last_modified = http.head(url).headers.get('last-modified', str(datetime.now()))
    entry = _try_lock_request(url, timeout)
    if type(entry) is not CacheItem or last_modified != entry.version:
        resp = http.urlopen(url=url, headers=headers, timeout=timeout)
        if 'b' in mode:
            entry = CacheItem(data=resp.bytes, version=last_modified)
        else:
            entry = CacheItem(data=resp.text, version=last_modified)
        add_item(key=url, item=entry)
    return _make_file_from_cache_entry(entry, url, mode)


def any_loc_open(url, timeout: int = http.DEFAULT_TIMEOUT, headers: Optional[dict] = None, mode='rb',
                 cache_remote_files=False) -> AnyFile:
    """Opens a file at the specified URL, whether local or remote.

    Parameters
    ----------
    url : str
        The file URL, formatted as either a local path or a standard URL (https://en.wikipedia.org/wiki/URL).
    timeout : int
        The timeout duration in seconds for remote files (default: 60 seconds).
    headers : Optional[dict]
        Optional HTTP headers to include when requesting remote files.
    mode : str
        The file open mode. Only 'r' or 'rb' are supported.
    cache_remote_files : bool
        Determines whether remote files are stored in the Speasy cache for future requests. Files are only downloaded
        if they have changed (based on the 'last-modified' header field).

    Returns
    -------
    AnyFile
        The opened file object.

    """
    if is_local_file(url):
        return AnyFile(url, open(url.replace('file://', ''), mode=mode))
    else:
        if cache_remote_files:
            return _cached_get_remote_file(url, timeout=timeout, headers=headers, mode=mode)
        else:
            return _remote_open(url, timeout=timeout, headers=headers, mode=mode)


def _list_local_files(path: str) -> List[str]:
    return os.listdir(path)


def _make_remote_files_relative(ref_path: str, path: str):
    if path.startswith('/') and len(ref_path) > 1:
        return path.removeprefix(ref_path)
    return path


@CacheCall(cache_retention=timedelta(hours=12), is_pure=True)
def _list_remote_files(url: str) -> List[str]:
    if not url.endswith('/'):
        url += '/'
    response = http.get(url)
    if response.ok:
        path = extract_path(url)
        return list(map(lambda f: _make_remote_files_relative(path, f), _HREF_REGEX.findall(response.text)))
    return []


def list_files(url: str, file_regex: Union[re.Pattern, str], disable_cache=False, force_refresh=False) -> List[str]:
    """Lists files that match the specified regex pattern either from a web page generated by Apache mod_dir or
    equivalent, or from a local directory.

    Parameters
    ----------
    url : str
        The URL or local path to scan.
    file_regex : re.Pattern or str
        The regular expression pattern used to filter files.
    disable_cache : bool
        Determines whether the cache is disabled for remote file listings.
    force_refresh : bool
        Forces a refresh of the cache for remote file listings.

    Returns
    -------
    List[str]
        A list of files that match the specified regex pattern, either from a remote source or a local directory.
    """
    if type(file_regex) is str:
        file_regex = re.compile(file_regex)
    if is_local_file(url):
        files = _list_local_files(url.replace('file://', ''))
    else:
        files = _list_remote_files(url, disable_cache=disable_cache, force_refresh=force_refresh)
    return list(filter(file_regex.match, files))
